from itertools import chain
import re
from shlex import quote

from log import log


ORDER = 'kegncpamdxytzw~'
FIXED_ATTRS = ('word', 'lemma', 'tag')


class Attrs(dict):  # TODO: Symbol? (Token?)
    def __str__(self, show_tag=False):
        if 'structural' in self:
            return self.stringify_structural()
        attrs = dict(self)
        attrs.pop('line', None)
        phrase_head = attrs.pop('phrase', None)
        phrase_head = [(phrase_head, None)] if phrase_head else []
        optional = '?' if attrs.pop('optional', None) else ''

        if not show_tag:
            _ = attrs.pop('tag', None)
        fixed_attrs = [(attr, attrs[attr]) for attr in FIXED_ATTRS
                       if attr in attrs]
        short_attrs = [(attr, attrs[attr]) for attr in ORDER if attr in attrs]
        other_attrs = sorted((attr, val) for (attr, val) in attrs.items()
                             if attr not in FIXED_ATTRS and (
                                len(attr) > 1 or attr not in ORDER))
        attrs = (attr + '=' + escape(val) if val is not None else attr
                 for attr, val in chain(phrase_head, fixed_attrs, short_attrs,
                                        other_attrs))
        return '({attrs}){token_is_optional}'.format(
            attrs=' '.join(attrs), token_is_optional=optional)

    def match(self, chart_edge):
        for attr, indexes in sorted(chart_edge.rule.attr_matches.items()):
            value = None
            for index in indexes:
                if index == -1:
                    continue
                if index < chart_edge.point:
                    if attr not in chart_edge.right[index]:
                        log.warning('             %s not in %s', attr,
                                    Attrs(chart_edge.right[index]))
                        continue
                    value = chart_edge.right[index][attr]
                elif value is not None and index == chart_edge.point:
                    if attr not in self:
                        log.warning('not present  %s %s', attr,
                                    str(self))
                    elif value is None:
                        log.debug('attr null    %s in %s', attr, self)
                    # TODO: přesunout funkci? (už se to vypisuje)
                    elif value != self[attr]:
                        log.error('mismatch     %s %s != %s in %s and %s',
                                  attr, self[attr], value, self, chart_edge)
                        return
        return True

    def stringify_structural(self):
        closing = '/' if self.get('closing') else ''
        tag = self.get('structural')
        empty = '/' if self.get('empty', '') else ''
        attrs = ''.join(
            ' {}="{}"'.format(attr, val) for attr, val in sorted(self.items())
             if attr not in ('line', 'closing', 'structural', 'empty'))
        return '<{closing}{tag}{attrs}{empty}>'.format(
            closing=closing, tag=tag, attrs=attrs, empty=empty)

    def vertical(self):
        return '{}\t{}\t{}'.format(self['word'], self['lemma'], self['tag'])


need_escaping = re.compile(r'[()=? \t\n]').search  # podle shlex._find_unsafe


def escape(s):  # začátek podle shlex.quote
    if not isinstance(s, str):
        s = str(s)
    if not s:
        return "''"
    elif need_escaping(s) is None:
        return s
    elif "'" not in s:
        return "'" + s + "'"
    elif '"' not in s:
        return '"' + s + '"'
    else:
        raise ValueError("Can't escape, the string includes both quotes: " + s)
