from attrs import Attrs
from log import log
from rule import Rule


class Edge(Rule):
    def __init__(self, predicted_from=None, predicted_by_rule=None,
                 extended_from=None, extended_using=None, tokens=None):
        self.predicted_from = predicted_from
        self.extended_from = extended_from
        # self.parent_position = None
        if extended_from:
            self.predicted_from = extended_from.predicted_from
            self.rule = extended_from.rule
            self.begin = extended_from.begin
            self.children = extended_from.children
            self.point = extended_from.point + 1
            self.left = Attrs(extended_from.left)
            self.right = [Attrs(token) for token in extended_from.right]
            if extended_using:
                self.end = extended_using.end
                self.real_offset = (extended_from.real_offset +
                                    extended_using.end - extended_using.begin)
                # log.critical('child        %s %s', extended_from.point,
                #              extended_using)
                self.children[extended_from.point] = extended_using
            elif tokens:  # terminal rule
                self.end = extended_from.end + 1
                self.real_offset = extended_from.real_offset + 1
                self.right[extended_from.point].update(
                    tokens[extended_from.begin + extended_from.real_offset])
                # nonterminal = self.rule.right[extended_from.point]
                # self.children[extended_from.point] = Edge(
                #     predicted_by_rule=Rule(
                #         left=self.right[extended_from.point],
                #         right=[self.right[extended_from.point]]))
            # self.parent_position = extended_from.parent_position
            # self.children[self.point]
        elif predicted_by_rule:
            self.rule = predicted_by_rule
            self.point = 0
            self.left = Attrs(self.rule.left)
            self.right = [Attrs(token) for token in self.rule.right]
            self.real_offset = 0
            if predicted_from:
                self.begin = predicted_from.end
                self.end = predicted_from.end
                # self.parent_position = predicted_from.point
                self._predict_attributes(predicted_from.next_token)
            else:
                self.begin = 0
                self.end = 0
        if not extended_from:
            self.children = [None] * len(self.right)
        self.complete = self.point == len(self.right)

    # můžou mít stejné pravidlo, stejný začátek i konec i tečku,
    # ale mohly vzniknout z jiných pravidel (ovšem i dřívějších),
    # takže porovnávat (i) rodiče – ale to už stačí, protože jejich
    # historie bude unikátní
    def __eq__(self, other):
        return (self.rule is other.rule and self.begin == other.begin and
                self.end == other.end and self.point == other.point)

    # def __hash__(self):
    #     return id(self.rule) ^ self.begin ^ self.end ^ self.point

    @property
    def next_token(self):
        return None if self.complete else self.right[self.point]

    @property
    def next_rule_token(self):
        return None if self.complete else self.rule.right[self.point]

    def copy_attributes(self, update=None):
        copy_from = self.point - 1
        if update:
            current = self.right[copy_from]
            for attr, value in sorted(update.items()):
                if value is None:
                    log.warning('null value   %s', attr)
                    continue  # asi by se nemělo kopírovat
                elif attr not in current or current[attr] is None:
                    log.debug('move         %s=%s → %s', attr, value, current)
                    current[attr] = value
                elif current[attr] != value:
                    log.warning('update       %s=%s → %s', attr, value,
                                current)
                    current[attr] = value

        for attr, indexes in sorted(self.rule.attr_matches.items()):
            if copy_from not in indexes:
                continue  # odsud se ani nemá kopírovat
            elif attr not in self.right[copy_from]:
                log.warning('no attr      %s %s', attr, self.right[copy_from])
                continue  # není co kopírovat

            value = self.right[copy_from][attr]
            if value is None:
                log.warning('null value   %s', attr)
                continue  # asi by se nemělo kopírovat

            if indexes[0] == -1:  # kopírování do hlavy
                log.debug('copy         %s=%s → head', attr, value)
                self.left[attr] = value

            # kopíruje se jen na členy vpravo od právě vloženého
            for index in indexes[indexes.index(copy_from)+1:]:
                log.debug('copy         %s=%s → %s', attr, value,
                          self.right[index])
                self.right[index][attr] = value

    def _predict_attributes(self, attrs):
        for attr, value in self.rule.left.items():
            if value is None and attr in attrs:
                self.left[attr] = attrs[attr]
                # log.debug('copy         %s=%s → head', attr, attrs[attr])

        for attr, match_on_indexes in sorted(self.rule.attr_matches.items()):
            if -1 in match_on_indexes and attr in self.left:
                value = self.left[attr]
                for index in match_on_indexes[1:]:
                    self.right[index][attr] = value
                    # log.debug('copy         head → %s=%s', attr, value)

    def update_parent(self):
        if not self.predicted_from:
            return
        self.predicted_from.children[self.parent_position] = self

    def add_child(self, edge):
        # log.critical('child        %s %s: %s', self.point, self, edge)
        self.children[self.point] = edge

    def _update_parent(self):
        originally_predicted = self.extended_from
        while originally_predicted.extended_from:
            originally_predicted = originally_predicted.extended_from
        # log.critical('original     %s', self)
        # log.critical('             %s', originally_predicted)
        parent = originally_predicted.predicted_from
        if not parent:
            log.critical('no parent    %s', self)
            log.critical('             %s', originally_predicted)
            return
        siblings = parent.children
        for index, edge in enumerate(siblings):
            if edge is originally_predicted:
                siblings[index] = self
                break
